""""
Main 21 cm forest class
Author: Pablo Villanueva Domingo
Last update: September 2022
"""

import numpy as np
import os
from tqdm import tqdm
from scipy import integrate, interpolate

from hayashi.constants import *
from hayashi import cosmo
from hayashi import subhalos
from hayashi import lss

#---------------------------
# 21cm Forest general class
#---------------------------
class Forest():
    def __init__(self,
                 z,                         # z: redshift of evaluation
                 Tk,                        # Tk: gas temperature of the IGM at that redshift
                 use_subhalos = True,       # use_subhalos: 1 for including the subhalo contribution, 0 otherwise
                 dndlnM = lss.dndlnM_CDM, # dndlnM: halo mass function for host halos (default: Sheth-Thormen)
                 sub_distr = "nfw",         # sub_distr: distribution of subhalos within the halo. Choose between "nfw" for the NFW profile, and "uni" for uniform distribution
                 tidal = False             # tidal: if True, use tidal disruption in outer parts of subhalos
                 ):

        # Redshift
        self.z = z
        # Temperature of the IGM at redshift z
        self.Tk = Tk
        # Flag, =True to take into account subhalo contribution
        self.use_subhalos = use_subhalos
        # Halo mass function
        self.dndlnM = dndlnM
        # Subhalo distribution within the host halo
        self.sub_distr = sub_distr
        # Consider tidal disruption in outer parts of subhalos
        self.tidal = tidal

        #--- Number of bins ---#
        # Number of bins of halo masses
        self.nummass = 60
        # Number of bins of the impact parameter
        self.numalf = 60
        # Number of bins of optical depth values
        self.numtau = 200

        # Jeans mass
        self.MJ = cosmo.MJeans(self.z,self.Tk)
        # Maximum mass for minihalos, taken as the minimum mass to host star formation, corresponding to a given virial temperature of 10^4 K
        self.Mmax = cosmo.Mmin(1.e4,z)

        # Mass, impact parameter and optical depth arrays
        self.logMvec = np.linspace(np.log(cosmo.MJeans(z,cosmo.Tk_ad(z))),np.log(self.Mmax),num=self.nummass)
        #self.logimpparam = np.linspace(np.log(1.e-5),0.,num=self.numalf)
        self.Mvec = np.exp(self.logMvec)
        #self.impparam = np.exp(self.logimpparam)   # impact parameter normalized to virial radius
        self.impparam = np.linspace(0.,1.,num=self.numalf)
        self.tauvec = np.logspace(-3,0,num=self.numtau)

        # Tidal radius: td_fac*r_s
        self.td_fac = 0.77  # value from astro-ph/0203004

        # Optical depth array (shape: tau[mass, impact param])
        self.tau_host_file = "Data/tau_z_{:.1f}".format(z)+"_nummass_"+str(self.nummass)+"_numalf_"+str(self.numalf)+".npy"
        if os.path.exists(self.tau_host_file):
            self.tau_host_arr = np.load(self.tau_host_file)
        else:
            print("Optical depth file for z="+str(z)+" does not exist. Generating...")
            self.tau_host_arr = self.get_optical_depth()

        # Subhalo contribution to optical depth array (shape: tau[mass, impact param])
        if self.use_subhalos:
            self.tau_sub_file = "Data/tau_sub_z_{:.1f}_Tk_{:.3e}".format(z,Tk)+"_"+self.sub_distr+"_tidal_"+str(int(self.tidal))+"_nummass_"+str(self.nummass)+"_numalf_"+str(self.numalf)+".npy"
            if os.path.exists(self.tau_sub_file):
                self.tau_sub_arr = np.load(self.tau_sub_file)
            else:
                print("Optical depth subhalo contribution file for z="+str(z)+", Tk={:.3e}".format(Tk)+" K does not exist. Generating...")
                self.tau_sub_arr = self.get_optical_depth_subhalos()
        else:
            self.tau_sub_arr = np.zeros((self.nummass,self.numalf))

        # Total optical depth
        self.tau_tot = self.tau_host_arr + self.tau_sub_arr

        # Boost factor for the optical depth
        self.boost_factor_arr = np.where(self.tau_host_arr==0., 0., self.tau_sub_arr/self.tau_host_arr)

        # Maximum impact parameter matrix (shape: tau[tau, mass])
        self.max_impact_arr = self.max_impact_param()

        # Create 2d interpolations of the optical depths, boost factor and maximum impact parameter
        # ..._int_logM expects the logarithm of mass as input, while ..._int expects as input the mass
        self.tau_host_int_logM = interpolate.interp2d(self.logMvec, self.impparam, np.transpose(self.tau_host_arr),kind="linear")
        self.tau_host_int = lambda M, al: self.tau_host_int_logM(np.log(M),al)
        self.tau_sub_int_logM = interpolate.interp2d(self.logMvec, self.impparam, np.transpose(self.tau_sub_arr),kind="linear")
        self.tau_sub_int = lambda M, al: self.tau_sub_int_logM(np.log(M),al)
        self.tau_tot_int = lambda M, al: self.tau_host_int(M, al) + self.tau_sub_int(M, al)
        self.boost_factor_int_logM = interpolate.interp2d(self.logMvec, self.impparam, np.transpose(self.boost_factor_arr),kind="linear")
        self.boost_factor_int = lambda M, al: self.boost_factor_int_logM(np.log(M),al)
        self.max_impact_int_log = interpolate.interp2d(np.log(self.tauvec), self.logMvec, np.transpose(self.max_impact_arr),kind="linear")
        self.max_impact_int = lambda tau, M: self.max_impact_int_log(np.log(tau),np.log(M))

    # Method to compute the optical depth array
    def get_optical_depth(self):

        tau_arr = np.zeros((self.nummass,self.numalf))

        pbar = tqdm(self.Mvec, total=self.Mvec.shape[0], position=0, leave=True, desc=f"Generating optical depth file")

        for iM, M in enumerate(pbar):
            for ia, impa in enumerate(self.impparam):

                tau_arr[iM,ia] = cosmo.OptDepth21(M,self.z,cosmo.con(M,self.z),impa*cosmo.Rvir(M,self.z))

        np.save(self.tau_host_file,tau_arr)

        return tau_arr

    # Method to compute the subhalo contribution to the optical depth array
    def get_optical_depth_subhalos(self):

        # Average tau
        tau_av = np.zeros(self.nummass)
        for iM, Msub in enumerate(self.Mvec):
            if self.tidal:
                stepfactor = np.heaviside(self.td_fac/cosmo.con(Msub,self.z) - self.impparam,1.)    # Max alpha: td_fac*rs (impparam is given as impact param normalized to Rvir, so a concentration factor appears)
            else:
                stepfactor = 1.
            tau_av[iM] = 2.*integrate.simps(self.tau_host_arr[iM,:]*self.impparam*stepfactor, self.impparam)

        tau_sub_arr = np.zeros((self.nummass,self.numalf))

        pbar = tqdm(self.Mvec, total=self.Mvec.shape[0], position=0, leave=True, desc=f"Generating optical depth subhalo contribution file")

        # For each host halo mass M
        for iM, Mhost in enumerate(pbar):
            if iM>0:
                for ia, impa in enumerate(self.impparam):

                    # Subhalo masses are those halo masses below the host halo mass
                    Msubs = self.Mvec[:iM]
                    avopdeps = tau_av[:iM]
                    # Number of subhalos array for each subhalo mass
                    numsubsalp = subhalos.num_subhalos_alpha(Msubs,self.z,cosmo.con(Mhost,self.z),impa*cosmo.Rvir(Mhost,self.z),Mhost,self.sub_distr)
                    #numsubsalp = np.array( [ subhalos.num_subhalos_alpha(Msub,self.z,cosmo.con(Mhost,self.z),impa*cosmo.Rvir(Mhost,self.z),Mhost,self.sub_distr) for Msub in Msubs ] )
                    # Write Jeans mass lower bound as a heaviside step function
                    stepfactor = np.heaviside(Msubs - self.MJ,1.)
                    tau_sub_arr[iM, ia] = integrate.simps(avopdeps*numsubsalp*stepfactor, Msubs)

        np.save(self.tau_sub_file,tau_sub_arr)

        return tau_sub_arr

    # Compute the maximum impact parameter for a given optical depth and mass
    def max_impact_param(self):

        max_impact_arr = np.zeros((self.numtau,self.nummass))

        for it, tau in enumerate(self.tauvec):
            maximpactvec = []
            for iM, M in enumerate(self.Mvec):
                maximpact = 0.
                for ia, impa in enumerate(self.impparam):
                    opdep = self.tau_tot[iM,ia]
                    if opdep>=tau:
                        maximpact = impa
                max_impact_arr[it, iM] = cosmo.Rvir(M,self.z)*maximpact

        return np.array(max_impact_arr)

    # Get the cumulative number of absorbers dN/dz and its logarithmic derivative taud^2N/dz/dtau
    def num_absorbers(self):

        cumulvec = np.zeros(self.numtau)
        prefactor = (1.+self.z)**2.*cosmo.drdz(self.z)/1.e6/(MpcToCm/hlittle)
        dlogtau = np.log(self.tauvec[1])-np.log(self.tauvec[0])

        for tt, tau in enumerate(self.tauvec):
            maximpactvec = self.max_impact_arr[tt]

            stepfactor = np.heaviside(self.logMvec - np.log(self.MJ),1.)
            intcumul = integrate.simps( self.dndlnM(np.exp(self.logMvec),self.z)*np.pi*maximpactvec**2.*stepfactor, self.logMvec )

            cumulvec[tt] = intcumul*prefactor

        # Compute the derivative of the cumulative number of absorbers with respect to tau
        dercumulvec = (cumulvec[1:] - cumulvec[:-1])/dlogtau

        self.cumulvec = cumulvec
        self.dercumulvec = dercumulvec

        return cumulvec, dercumulvec



""""
Subhalos module
Includes the relevant routines for computing the subhalo contribution to the optical depth
Pablo Villanueva Domingo, Kenji Kadota
Started: October 2021
Last update: September 2022
"""


from hayashi.cosmo import *

# Set random seed for Monte Carlo integration
np.random.seed(0)

# Number of samples drawn within the domain for the Monte Carlo integraton. The bigger the more accurate but slower
Nsamples = 10000

# Subhalo mass function
# Masses Msub and Mhost in Msun     CHECK h factors!
# submf: Subhalo mass function parameterization
# Choose between "Ando19" (1903.11427) and "Dooley17" (1610.00708)
def subhalo_mf(Msub, z, Mhost, submf = "Ando19"):

    # Eq. 2 1610.00708
    if submf=="Dooley17":
        K0, chi = 1.88e-3, 1.87 # K0 in Msun^-1
        return K0*(Msub)**(-chi)*Mhost
    # Appendix A of 1903.11427
    elif submf=="Ando19":
        a_h = 6.0e-4*(1.+z)*Mhost*(2.-0.08*np.log10(Mhost))**2.*(np.log10(Mhost/1.e-5))**(-0.1)
        b_h = 8.0e-5*Mhost*(2.-0.08*np.log10(Mhost))*(np.log10(Mhost/1.e-5))**(-0.08*z)*(np.log10(Mhost/1.e-8))**(-1.)*(np.log10(Mhost/1.e18))**(2.)
        mc = 0.05*(1.+z)*Mhost
        alpha_h = 0.2 + 0.02*z
        beta_h = 3.
        return (a_h + b_h*Msub**alpha_h)/Msub**2.*np.exp(-(Msub/mc)**beta_h)

# Normalized density profile
def w_dens(Mhost,z,y,r,sub_distr):
    if sub_distr=="nfw":
        return rho_nfw(Mhost,z,y,r)/Mhost
    elif sub_distr=="uni":
        return rho_c_z(z)*Deltac(z)/Mhost

# Number of subhalos within a cylinder at a given impact parameter from the center of the host halo, using Monte Carlo integration
# Msub: host halo mass
# z: redshift
# y: concentration of the host halo
# al: impact parameter of the host halo (in kpc/h)
# Mhost: host halo mass
def num_subhalos_alpha(Msub,z,y,al,Mhost,sub_distr):

    Rhost, Rsub = Rvir(Mhost,z), Rvir(Msub,z)
    alplus, almin = min(al+Rsub,Rhost), max(al-Rsub,0.)
    unitfact = (1.e-3*MpcToCm)**3./Msun

    if Rhost < (al + Rsub):
        Nsub = 0.0

    else:

        zzmax = np.sqrt(Rhost ** 2 - (al + Rsub) ** 2)
        rrmax = Rsub
        thetamax = 2.0 * np.pi
        zzmin = 0.0
        rrmin = 0.0
        thetamin = 0.0

        domain = (thetamax - thetamin) * (zzmax - zzmin) * (rrmax - rrmin)

        rr = rrmax*np.sqrt(np.random.uniform(low=0., high=1., size=Nsamples))  # N values uniformly drawn from a to b
        theta = np.random.uniform(low=thetamin, high=thetamax, size=Nsamples)  # N values uniformly drawn from c to d
        zz = np.random.uniform(low=zzmin, high=zzmax, size=Nsamples)  # N values uniformly drawn from 0 to 2Pi

        rad = np.sqrt(rr ** 2 * np.sin(theta) ** 2 + (al + rr * np.cos(theta)) ** 2 + zz ** 2)

        integrand = w_dens(Mhost,z,y,rad,sub_distr)*rr

        Nsub = 2 * (np.sum(integrand) * domain) / Nsamples * subhalo_mf(Msub, z, Mhost) * unitfact

    return Nsub

num_subhalos_alpha = np.vectorize(num_subhalos_alpha)



""""
Physical and cosmological constants
Author: Pablo Villanueva Domingo
Started: October 2018
Last update: March 2021
"""

import numpy as np

#----- Constants -----#

# Misc
hplanck = 6.62606896e-27    #  erg s
Msun = 1.989e33 # g
G = 6.67259e-8  # cm^3 g^-1 s^-2
m_p = 1.6726231e-24 # g
m_ec2 = 0.5109989e6 # eV
m_pc2 = 0.938272e9 # eV
sigmaT = 6.6524e-25 #cm^2
c = 29979245800.0 #cm/s
kB = 8.6173324e-5 # eV/K    #1.380658e-16 # erg/K
alphafine = 1./(137.035999)
e2 = alphafine*hplanck*c/2./np.pi   # erg * cm
hnu_0 = 13.6    # eV
sigmaH = 7.91e-18   # ion cross section, cm^2

# Units
nu_over_ev = 1.60217646e-12/hplanck # =2.417e14, 1 eV = nu_over_ev Hz
year_sec = 3.154e7 # 1yr in seconds
MpcToCm = 3.086e24 # 1 Mpc = MpcToCm cm
eVtoerg = 1.60218e-12 # 1 eV = eVtoerg erg
kelvintoerg = 1.38064878066852e-16 # 1 kelvin = kelvintoerg erg
kelvintoeV = kelvintoerg/eVtoerg   # 1 kelvin = kelvintoeV eV

# Cosmo
G_N = 6.67259e-8 # cm^3 g^-1 s^-2
hlittle = 0.678 #0.678
H0 = hlittle*3.2407e-18 # s^-1
Omega_b = 0.02226/hlittle**2.
Omega_m = 0.308 #0.3158
Omega_dm = Omega_m - Omega_b
Omega_l = 0.68 #1.-Omega_m
Y_He = 0.245    # mass fraction
mu = 1./(1.-3.*Y_He/4.) # mean molecular weight, mu=1.22 for neutral H + He
rho_c = 3.*H0**2./(8.*np.pi*G) # g cm^-3
n0_h = rho_c*Omega_b*(1.-Y_He)/m_p   # #/cm^3
n0_he = rho_c*Omega_b*Y_He/(4.*m_p)  # #/cm^3
n0_b = n0_h + n0_he     # n0_b = rho_c/(mu*m_p)*Omega_b
Tcmb0 = 2.7255 # K
rho_cmb0 = 8.*np.pi**5.*(kB*Tcmb0)**4./15./(hplanck*c/eVtoerg)**3. # eV/cm^3
LEdd = 4.*np.pi*G_N*Msun*c*m_p/sigmaT   # LEdd/(M/Msun) in erg/s (7.84635981059e+46 keV/s)
zdec = 150.*(Omega_b*hlittle**2./0.022)**(2./5.) - 1.   # z decoupling of Tk


# 21 cm and Ly alpha
Tstar = 0.0682   # h\nu_21 [K]
A10 = 2.85e-15  # hyperfine Einstein coefficient s^-1
nu21 = 1420.4e6    # hyperfine frecuency Hz
f12 = 0.416     # f12 for Lymanalpha
nualpha = 2.47e15   # Hz
Aalpha = 6.25e8 # s^-1
tauGP = 3.*c**3.*Aalpha*n0_h/(8.*np.pi*nualpha**3.*H0*np.sqrt(Omega_m))*10.**(3./2.)    # tau_GP at z=10 and x_HI=1, wrong value? not used

"""
#----- Precision Constants -----#

# Number of bins of halo masses
nummass = 100
# Number of bins of optical depth values
numtau = 300
# Number of bins of the impact parameter
numalf = 100
# log cutoff in integrals
cutoff = 10.
# num of k values in k integrals, set 1000 for dndlnM plot
num_k = 1000

# For including the subhalo contribution, choose 1, 0 otherwise
use_subhalos = 1
"""



""""
Cosmological functions
Pablo Villanueva Domingo
Started: October 2018
Last update: April 2021
"""


from scipy import integrate
from hayashi.constants import *

# Employ the Colossus package for some cosmological computations:
# https://bdiemer.bitbucket.io/colossus/tutorials.html
from colossus.cosmology import cosmology
from colossus.halo.concentration import concentration

# Define a cosmological model with Colossus
cosmomodel = cosmology.setCosmology('planck18')
rho_c_Mpc = cosmomodel.rho_c(0)*1e9   # in M_sun h^2 Mpc^-3

#--- COSMO ---#

# Hubble function, s^-1
def Hz(z):
    return H0*np.sqrt(Omega_m*(1.+z)**3. + Omega_l)

# dt/dz, in seconds
def dtdz(z):
    return -1./Hz(z)/(1.+z)

# dr/dz, in cm
def drdz(z):
    return -c*dtdz(z)*(1.+z)

# Adiabatic temperature, in K
def Tk_ad(z):
    return Tcmb0*(1.+zdec)*((1.+z)/(1.+zdec))**2.

# Jeans mass, Msun/h
def MJeans(z,T): # \simeq 3.58e5*(Tk_ad(z)/(1.+z))**(3./2.) Msun/hlittle ?
    lambdaJ = (5.*np.pi*T*kelvintoerg/(3.*G_N*rho_m_z(z)*mu*m_p))**(1./2.)
    #return 4.*np.pi*rho_m_z(z)/3.*(lambdaJ)**(3.)/(Msun/hlittle)   # \simeq 4.96e5*(Tk/(1.+z))**(3./2.) Msun
    return 4.*np.pi*rho_m_z(z)/3.*(lambdaJ/2.)**(3.)/(Msun/hlittle)   # \simeq 6.21e4*(Tk/(1.+z))**(3./2.) Msun
    #return 4.*np.pi*rho_m_z(z)/3.*(5.*3.*T*kelvintoerg/(4.*np.pi*G_N*rho_m_z(z)*mu*m_p))**(3./2.)/(Msun/hlittle)

# Matter density, g cm^-3
def rho_m_z(z):
    return Omega_m*rho_c*(1.+z)**3.

# Critical density as a function of z, g cm^-3
def rho_c_z(z):
    return rho_c*( Omega_m*(1.+z)**3. + Omega_l )


#--- HALO VIRIALITIES ---#

# Omega_m*(1.+z)**3./(Omega_m*(1.+z)**3. + Omega_l)
def Omz(z):
    return rho_m_z(z)/rho_c_z(z)

# Virial overdensity
def Deltac(z):
    d = Omz(z) -1.
    return 18.*np.pi**2. + 82.*d - 39.*d**2.

# Virial radius, kpc/h, M in Msun/h, physical units
def Rvir(M,z):
    return 0.784*(M/1.e8)**(1./3.)*(Omega_m/Omz(z)*Deltac(z)/(18.*np.pi**2.))**(-1./3.)*((1.+z)/10.)**(-1.)

# Circular velocity, km/s, Vcirc = sqrt(G*M/rvir)
def Vcirc(M,z):
    return 23.4*(M/1.e8)**(1./3.)*(Omega_m/Omz(z)*Deltac(z)/(18.*np.pi**2.))**(1./6.)*((1.+z)/10.)**(1./2)

# Virial temperature, K
# Barkana & Loeb 2001 has factor 1.98, while Shimabukuro et al 2014 has 1.32, as factor 3/2 lower
def Tvir(M,z):
    return 1.98e4*(mu/0.6)*(M/1.e8)**(2./3.)*(Omega_m*Deltac(z)/(Omz(z)*18.*np.pi**2.))**(1./3.)*((1.+z)/10.)

# Virial mass from a given virial temperature, Msun/h, Tvir in K
def Mmin(Tvir,z):
    return 1.e8*(Omega_m*Deltac(z)/(Omz(z)*18.*np.pi**2.))**(-1./2.)*(Tvir/1.98e4)**(3./2.)*((1.+z)/10.)**(-3./2.)

# Halo concentration parameter
def con(M,z,model="ishiyama21"):
    if model=="Bullock2001":      # Employs fit from Bullock et al. 2001, to match at z=0
        return 9./(1.+z)*(M/1.5e13)**(-0.13)
    else:     # Employs fit from colossus. Default here is taken as "ishiyama21"
        return concentration(M, "vir", z=z, model=model)


#--- DENSITY PROFILES ---#

# Integral appearing in NFW profile
def Fint(y):
    return np.log(1.+y)-y/(1.+y)

# Squared escape velocity, (km/s)^2
def Vesc2(M,z,y,r):
    x = r/Rvir(M,z)
    return 2.*Vcirc(M,z)**2.*( Fint(y*x) + y*x/(1.+y*x) )/(x*Fint(y))

# Central gas density, g/cm^3 . It fulfills M_gas = (Omega_b/Omega_m)*M_nfw
def rho_gas_central(z,y):
    A = 2.*y/Fint(y)
    lntvec = np.linspace(np.log(1.e-10),np.log(y))
    integral = integrate.simps((1.+np.exp(lntvec))**(A/np.exp(lntvec))*np.exp(lntvec)**3.,lntvec)
    return (Omega_b/Omega_m)*rho_c_z(z)*Deltac(z)/3.*y**3.*np.exp(A)/integral

# Gas density profile, obtained assuming an isothermal halo in hydrostatic equilibrium, g/cm^3
def rho_gas(M,z,y,r):
    Vesc2_0 = 2.*Vcirc(M,z)**2.*y/Fint(y)
    return rho_gas_central(z,y)*np.exp(-mu*m_p*(Vesc2_0-Vesc2(M,z,y,r))*1.e10/(2.*Tvir(M,z)*kelvintoerg))

# NFW density profile, g/cm^3
def rho_nfw(M,z,y,r):
    rho_0 = rho_c_z(z)*Deltac(z)/3.*y**3./Fint(y)
    x = y*r/Rvir(M,z)
    return rho_0/x/(1.+x)**2.


#--- 21 CM ---#

# Scattering rate for Hydrogen-Hydrogen collisions
def kappa_HH(T):
    if T>10:
        kh = 3.1e-11*T**(0.357)*np.exp(-32./T)
    elif T<=10 and T>1:
        kh = 3.6e-16*T**(3.640)*np.exp(6.035/T)
    else:
        kh = 3.6e-16*1.**(3.640)*np.exp(6.035/1.)
    return kh

# Collision coupling coefficient y
def yColl(z,nH,T):
    return Tstar/T/A10*kappa_HH(T)*nH

# Spin temperature, K
def Ts(z,nH,Tk):
    yc = yColl(z,nH,Tk)
    return (Tcmb0*(1.+z) + yc*Tk)/(1. + yc)

# 21 cm optical depth
def OptDepth21(M,z,y,al):
    prefactor = 3.*hplanck*c**3.*A10/(32.*np.pi*nu21**2.)   # cm^3*erg/s
    RRmax = np.sqrt(1.-(al/Rvir(M,z))**2.)        # Maximum radius to integrate
    if RRmax==0.:   # avoid division by 0
        return 0.
    logRR = np.linspace(np.log(1.e-3),np.log(RRmax),100)
    RR = np.exp(logRR)
    rvec = np.sqrt(RR**2. + (al/Rvir(M,z))**2.)
    Tk = Tvir(M,z)
    nH = rho_gas(M,z,y,rvec*Rvir(M,z))*(1.-Y_He)/m_p
    Tspin = Ts(z,nH,Tk)
    b = np.sqrt(2.*Tk*kelvintoerg/m_p)      # cm/s
    nu = nu21/(1.+z)
    expdoppler = 1./np.sqrt(np.pi)/b        # Doppler exponential factor
    #expdoppler = 1./np.sqrt(np.pi)/b*np.exp(-(c*(nu-nu21)/nu21)**2./b**2.)
    integral = 2.*Rvir(M,z)*integrate.simps( RR*nH/Tspin*expdoppler, logRR) # 2 is 'cause goes from 0 to Rmax
    #integral = 2.*Rvir(M,z)*integrate.quad(lambda logR: np.exp(logR)*rho_gas(M,z,y,np.sqrt(np.exp(logR)*2.+(al/Rvir(M,z))**2.)*Rvir(M,z))*(1.-Y_He)/m_p/Ts(z,rho_gas(M,z,y,np.sqrt(np.exp(logR)*2.+(al/Rvir(M,z))**2.)*Rvir(M,z))*(1.-Y_He)/m_p,Tk)*expdoppler, logRR[0], logRR[-1])[0]
    return prefactor*integral/kelvintoerg*MpcToCm/1.e3/hlittle



""""
Large scale structure routines
Pablo Villanueva Domingo
Started: October 2018
Last update: September 2022
"""

from hayashi.cosmo import *
from colossus.lss import mass_function

# log cutoff in integrals
cutoff = 10.
# num of k values in k integrals, set 1000 for dndlnM plot
num_k = 1000

#--- Standard large scale structure ---#

# Radius conversion from a given mass
def MtoR(M):
    return (4.*np.pi/3.*Omega_m*rho_c_Mpc)**(-1./3.)*M**(1./3.)

# Top-hat window
def WindowTH(x):
    return 3*(np.sin(x)-x*np.cos(x))/x**3

# Logarithmic derivative of the top-hat window, dW/dlnR
def DerWindowTH(x):
    return (9*x*np.cos(x)+3*(x**2 - 3)*np.sin(x))/x**3

# Press-Schechter first crossing distribution
def f_ps(x):
    return np.sqrt(2.*x/np.pi)*np.exp(-x/2.)

# Sheth-Tormen first crossing distribution
def f_st(x):
    p = 0.3
    q = 0.707
    A = 0.3222
    return A*np.sqrt(2*q*x/np.pi)*( 1 + (q*x)**(-p) )*np.exp(-q*x/2)

# Halo mass function in LCDM cosmology, dndlnM, units of (h/Mpc)^3, using the Sheth-Tormen parameterization
def dndlnM_CDM(M,z):
    return mass_function.massFunction(M, z, q_out = "dndlnM", model = "sheth99")



""""
Non-LambdaCDM large scale structure (warm dark matter and primordial black holes)
Pablo Villanueva Domingo
Started: October 2018
Last update: September 2022
"""

from hayashi.lss import *


#--- Primordial Black Holes ---#

# Isocurvature power spectrum for PBHs at z=0, (Mpc/h)^3
def IsoPS(k,fpbh,Mpbh):
    onepluszeq = 2.4e4*Omega_m*hlittle**2
    keq = 0.073*Omega_m*hlittle    # h/Mpc
    if k > keq:
        return (9./4.)*onepluszeq**2.*fpbh*Mpbh*hlittle/((Omega_m-Omega_b)*rho_c_Mpc)
    else:
        return 0.

# Total power spectrum for PBH universes at z=0, (Mpc/h)^3
def PkPBH(k,fpbh,Mpbh):
    IsoPSvec = np.vectorize(IsoPS)
    return cosmomodel.matterPowerSpectrum(k) + IsoPSvec(k,fpbh,Mpbh)

# Variance of the linear field in PBH scenarios
def Sigma2_PBH(M,fpbh,Mpbh):
    R = MtoR(M)
    logk = np.linspace(-cutoff,cutoff,num=num_k)
    return integrate.simps((1./(2.*np.pi**2.))*np.exp(3.*logk)*PkPBH(np.exp(logk),fpbh,Mpbh)*WindowTH(np.exp(logk)*R)**2, logk )

# dsigma^2/dlnM in PBH scenarios
def DerSigma2_PBH(M,fpbh,Mpbh):
    R = MtoR(M)
    logk = np.linspace(-cutoff,cutoff,num=num_k)
    return integrate.simps((2./3.)*(1/(2.*np.pi**2.))*np.exp(3.*logk)*PkPBH(np.exp(logk),fpbh,Mpbh)*WindowTH(np.exp(logk)*R)*DerWindowTH(np.exp(logk)*R), logk )

# Halo mass function for PBH scenarios, dndlnM, units of (h/Mpc)^3
def dndlnM_PBH(M,z,fpbh,Mpbh):
    s2 = Sigma2_PBH(M,fpbh,Mpbh)
    deltac = 1.68647/cosmomodel.growthFactor(z)
    nu = (deltac**2/s2)
    dlogsigdlogm = np.abs(DerSigma2_PBH(M,fpbh,Mpbh)/(2.*s2))
    return Omega_m*rho_c_Mpc*f_st(nu)/M*dlogsigdlogm



#--- Warm Dark Matter ---#

# WDM transfer function
def Transfer2(k,m):
    nu = 1.12
    alpha = 0.049*m**(-1.11)
    return (1.+(alpha*k)**(2.*nu))**(-10./nu)

# WDM Power spectrum
def ps_wdm(k,m):
    return Transfer2(k,m)*cosmomodel.matterPowerSpectrum(k)

# Half-mode length
def lamb_hmm(m):
    nu = 1.12
    alpha = 0.049*m**(-1.11)
    return 2.*np.pi*alpha*(2.**(nu/5.) - 1.)**(-1./2./nu)

# Half-mode mass
def M_hmm(m):
    return (4.*np.pi/3.*Omega_m*rho_c_Mpc)*(lamb_hmm(m)/2.)**3.

# Variance of the linear field in WDM scenarios
def Sigma2_WDM(M,mwdm):
    R = MtoR(M)
    logk = np.linspace(-cutoff,cutoff,num=num_k)
    return integrate.simps((1./(2.*np.pi**2.))*np.exp(3.*logk)*ps_wdm(np.exp(logk),mwdm)*WindowTH(np.exp(logk)*R)**2, logk )
    #return integrate.quad(lambda logk: (1./(2.*np.pi**2.))*np.exp(3.*logk)*PkPBH(np.exp(logk),fpbh,Mpbh)*WindowTH(np.exp(logk)*R)**2, -cutoff, cutoff)[0]

# dsigma^2/dlnM in WDM scenarios
def DerSigma2_WDM(M,mwdm):   # dsigma^2/dlnM
    R = MtoR(M)
    logk = np.linspace(-cutoff,cutoff,num=num_k)
    return integrate.simps((2./3.)*(1/(2.*np.pi**2.))*np.exp(3.*logk)*ps_wdm(np.exp(logk),mwdm)*WindowTH(np.exp(logk)*R)*DerWindowTH(np.exp(logk)*R), logk )
    #return integrate.quad(lambda logk: (2./3.)*(1/(2.*np.pi**2))*np.exp(3.*logk)*PkPBH(np.exp(logk),fpbh,Mpbh)*WindowTH(np.exp(logk)*R)*DerWindowTH(np.exp(logk)*R), -cutoff, cutoff)[0]

def dndlnM_WDM(M,z,mwdm):  # dndlnM units of (h/Mpc)^3
    s2 = Sigma2_WDM(M,mwdm)
    deltac = 1.68647/cosmomodel.growthFactor(z)
    nu = (deltac**2/s2)
    dlogsigdlogm = np.abs(DerSigma2_WDM(M,mwdm)/(2.*s2))
    return Omega_m*rho_c_Mpc*f_st(nu)/M*dlogsigdlogm*(1. + 2.*M_hmm(mwdm)/M)**(-0.6)

# Vectorize some functions to allow them to take arrays as inputs
Sigma2_PBH = np.vectorize(Sigma2_PBH)
DerSigma2_PBH = np.vectorize(DerSigma2_PBH)
Sigma2_WDM = np.vectorize(Sigma2_WDM)
DerSigma2_WDM = np.vectorize(DerSigma2_WDM)



from setuptools import find_packages, setup

setup(
    name='hayashi',
    packages=find_packages(),
    version='1.1.0',
    description='Python library for computing the number of absorption features of the 21 cm forest in a semianalytic formalism.',
    author='Pablo Villanueva-Domingo',
    author_email='pablo.villanueva.domingo@gmail.com',
    install_requires=["numpy","scipy","tqdm","colossus"],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',  # License type
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)



""""
Utility functions
Pablo Villanueva Domingo
Started: October 2018
Last update: July 2022
"""

import numpy as np
import math
from scipy import interpolate
from hayashi.cosmo import Tk_ad
import os

# Create some folders for outputs if they are absent
for path in ["Plots", "Data"]:
        if not os.path.exists(path):
            os.mkdir(path)


#--- MISC ---#

# Write a number in Latex scientific notation
def scinot(x):
    exp = int(math.floor(math.log10(abs(x))))
    prefactor = x / 10**exp
    if exp==0:
        return r"{:.0f}".format(prefactor)
    if exp==1:
        return r"{:.0f}".format(prefactor*10.)
    elif prefactor == 1.:
        return r"$10^{"+str(exp)+"}$"
    else:
        return r"${:.0f}".format(prefactor)+" \\times 10^{"+str(exp)+"}$"


# Import the IGM temperature from interpolation
def GetTemperature(z, zetax):

    if zetax==0.:
        return Tk_ad(z)

    # Take temperature from file if it exists
    ThermalHisF = 'ThermalHistory_Xi_30_Tmin_1.000e+04_Rfmp_15_chiUV_{:.2e}_Nalpha_4.00e+03.dat'.format(zetax)
    file = np.loadtxt("21cmFiles/"+ThermalHisF,unpack=True)
    Temp = interpolate.interp1d(file[0], file[2], fill_value="extrapolate")
    Tk = Temp(z)
    return Tk
